// /client.js
 

var superagent = require('ti.superagent');

var Client = function Client() {
	/**
	 * The base path to put in front of every API call's (relative) path.
	 */
	this.basePath = 'https://10.129.59.159:8065/api'.replace(/\/+$/, '');
	this.authentications = {
		get: function( name ) { return this[name]; }, set: function( name, value ) { this[name] = value;},
		'OAuthImplicit': {type: 'oauth2', scopes: 'resource.WRITE,resource.READ', authorizationUrl: 'http://localhost:8089/api/oauth/authorize'}
	};
	
	/**
	 * The default HTTP headers to be included for all API calls.
	 */
	this.defaultHeaders = {};

	/**
	 * The default HTTP timeout for all API calls.
	 */
	this.timeout = 60000;

	/**
	 * By default, no caching is used.
	 */
	this.cachePolicy = Client.CachePolicy.NoCache;
};

// ----------- Caching ---------------
// Caching policy enum
Client.CachePolicy = {
	/**
	 * No caches are used or saved. This is the default caching policy for models.
	 */
	NoCache: 1,
	/**
	 * If a local cache of the results is available, it will be used; otherwise, the network will be used. Results will
	 * be cached.
	 */
	CacheElseNetwork: 2,
	/**
	 * If the network is available, it will be used for results; otherwise, the local cache will be used. Results will
	 * be cached.
	 */
	NetworkElseCache: 4,
	/**
	 * If a local cache of the result is available, it will be used immediately, and then the network will be used. The
	 * callback will, thus, be called twice. Results will be cached.
	 */
	CacheThenNetwork: 8,
	/**
	 * If a local cache of results is available, it will be used.
	 */
	CacheOnly: 16,
	/**
	 * If the network is available, it will be used for results. Results will be cached.
	 */
	NetworkOnly: 32
};

/**
 * Turns a url and method in to a hash that can uniquely and securely identify it.
 */
Client.prototype.getCacheHash = function (method, url, headers) {
	var key = method + ':' + url + '\n';
	// Add headers.
	for (var headerName in headers) {
		if (headers.hasOwnProperty(headerName)) {
			key += headerName + ':' + headers[headerName] + '\n';
		}
	}
	return Ti.Utils.sha256(key);
};

/**
 * Gets the cache db, ensuring it has the proper schema set up.
 * @returns {LRUCache}
 */
Client.prototype.getCacheDB = function () {
	if (this.cacheDB) {
		return Ti.Database.open(this.cacheDB);
	}
	this.cacheDB = 'studentcarCache';
	var db = Ti.Database.open(this.cacheDB);
	db.execute('CREATE TABLE IF NOT EXISTS cache(hash TEXT PRIMARY KEY, until INTEGER, result BLOB);');
	db.file.setRemoteBackup(false);
	return db;
};

/**
 * Checks the cache to see if results have been cached for the specific request.
 * @param hash A string that uniquely identifies the request.
 * @returns The cached results, or undefined.
 */
Client.prototype.readCache = function (hash) {
	var db = this.getCacheDB();
	db.execute('DELETE FROM cache WHERE until < ?', Date.now());
	var result = db.execute('SELECT result FROM cache WHERE hash = ? LIMIT 1', hash),
		retVal;
	if (result.isValidRow()) {
		retVal = JSON.parse(result.fieldByName('result'));
	}
	result.close();
	db.close();
	return retVal;
};

/**
 *
 * @param hash
 * @param duration
 * @param results
 */
Client.prototype.writeCache = function (hash, duration, results) {
	var db = this.getCacheDB();
	db.execute('INSERT OR REPLACE INTO cache (hash, until, result) VALUES (?, ?, ?);',
		hash,
		Date.now() + duration,
		JSON.stringify(results));
	db.close();
};

Client.prototype.paramToString = function paramToString(param) {
	if (param === null) {
		// return empty string for null and undefined
		return '';
	} else if (param instanceof Date) {
		return param.toJSON();
	} else {
		return param.toString();
	}
};

/**
 * Build full URL by appending the given path to base path and replacing
 * path parameter placeholders with parameter values.
 * NOTE: query parameters are not handled here.
 */
Client.prototype.buildUrl = function buildUrl(path, pathParams) {
	if (!path.match(/^\//)) {
		path = '/' + path;
	}
	var url = this.basePath + path;
	var _this = this;
	url = url.replace(/\{([\w-]+)\}/g, function(fullMatch, key) {
		var value;
		if (pathParams.hasOwnProperty(key)) {
			value = _this.paramToString(pathParams[key]);
		} else {
			value = fullMatch;
		}
		return encodeURIComponent(value);
	});
	return url;
};

/**
 * Check if the given MIME is a JSON MIME.
 * JSON MIME examples:
 *   application/json
 *   application/json; charset=UTF8
 *   APPLICATION/JSON
 */
Client.prototype.isJsonMime = function isJsonMime(mime) {
	return Boolean(mime !== null && mime.match(/^application\/json(;.*)?$/i));
};

/**
 * Choose a MIME from the given MIMEs with JSON preferred,
 * i.e. return JSON if included, otherwise return the first one.
 */
Client.prototype.jsonPreferredMime = function jsonPreferredMime(mimes) {
	var len = mimes.length;
	for (var i = 0; i < len; i++) {
		if (this.isJsonMime(mimes[i])) {
			return mimes[i];
		}
	}
	return mimes[0];
};

/**
 * Check if the given parameter value is like file content.
 */
Client.prototype.isFileParam = function isFileParam(param) {
	// Buffer in Titanium
	if (typeof Ti.Buffer === 'function' && param instanceof Ti.Buffer) {
		return true;
	}
	// Ti.Blob in Titanium
    // Ti.Blob doesn't handle instanceof checks properly
    // if (typeof Ti.Blob === 'function' && param instanceof Ti.Blob) {
	if (param.apiName === 'Ti.Blob') {
		return true;
	}
	// File in Titanium
	if (typeof Ti.Filesystem.File === 'function' && param instanceof Ti.Filesystem.File) {
		return true;
	}
	return false;
};

/**
 * Normalize parameters values:
 *   remove nils,
 *   keep files and arrays,
 *   format to string with `paramToString` for other cases.
 */
Client.prototype.normalizeParams = function normalizeParams(params) {
	var newParams = {};
	for (var key in params) {
		if (params.hasOwnProperty(key) && params[key] !== null) {
			var value = params[key];
			if (this.isFileParam(value) || Array.isArray(value)) {
				newParams[key] = value;
			} else {
				newParams[key] = this.paramToString(value);
			}
		}
	}
	return newParams;
};

/**
 * Build parameter value according to the given collection format.
 * @param {String} collectionFormat one of 'csv', 'ssv', 'tsv', 'pipes' and 'multi'
 */
Client.prototype.buildCollectionParam = function buildCollectionParam(param, collectionFormat) {
	if (param === null) {
		return null;
	}
	switch (collectionFormat) {
		case 'csv':
			return param.map(this.paramToString).join(',');
		case 'ssv':
			return param.map(this.paramToString).join(' ');
		case 'tsv':
			return param.map(this.paramToString).join('\t');
		case 'pipes':
			return param.map(this.paramToString).join('|');
		case 'multi':
			// FIXME How do we handle this?
			// return the array directly as Superagent will handle it as expected
			return param.map(this.paramToString);
		default:
			throw new Error('Unknown collection format: ' + collectionFormat);
	}
};

Client.prototype.applyAuthToRequest = function applyAuthToRequest(request, authNames) {
	var _this = this;
	authNames.forEach(function(authName) {
		var auth = _this.authentications[authName];
		switch (auth.type) {
			case 'basic':
				if (auth.username || auth.password) {
					request.auth(auth.username || '', auth.password || '');
				}
				break;
			case 'apiKey':
				if (auth.apiKey) {
					var data = {};
					if (auth.apiKeyPrefix) {
						data[auth.name] = auth.apiKeyPrefix + ' ' + auth.apiKey;
					} else {
						data[auth.name] = auth.apiKey;
					}
					if (auth.in === 'header') {
						request.set(data);
					} else {
						request.query(data);
					}
				}
				break;
			case 'oauth2':
				// TODO Automatically go through the flow for oauth if we need to grab the token!
				if (auth.accessToken) {
					request.set({'Authorization': auth.tokenType + ' ' + auth.accessToken});
				}
				break;
			default:
				throw new Error('Unknown authentication type: ' + auth.type);
		}
	});
};

Client.prototype.deserialize = function deserialize(response, returnType) {
	if (response === null || returnType === null) {
		 return null;
	}
	// Rely on Superagent for parsing response body.
	// See http://visionmedia.github.io/superagent/#parsing-response-bodies
	var data = response.body;
	if (data === null) {
		// Superagent does not always produce a body; use the unparsed response
		// as a fallback
		data = response.text;
	}
	return Client.convertToType(data, returnType);
};

Client.prototype.callApi = function callApi(path, httpMethod, pathParams,
 queryParams, headerParams, formParams, bodyParam, authNames, contentTypes,
 accepts, returnType, callback) {
	var _this = this;
	var url = this.buildUrl(path, pathParams);
	var request = superagent(httpMethod.toUpperCase(), url);

	// apply authentications
	this.applyAuthToRequest(request, authNames);

	// set query parameters
	request.query(this.normalizeParams(queryParams));

	// set header parameters
	request.set(this.defaultHeaders).set(this.normalizeParams(headerParams));

	// TODO Move caching code here. We need to take into account the URL, plus query params, plus headers!
	// right now it assumes headers live on config, and that URl already has the query params embedded!

	// set request timeout
	request.timeout(this.timeout);

	var contentType = this.jsonPreferredMime(contentTypes);
	if (contentType) {
		request.type(contentType);
	} else if (!request.header['Content-Type']) {
		request.type('application/json');
	}

	if (contentType === 'application/x-www-form-urlencoded') {
		request.send(this.normalizeParams(formParams));
	} else if (contentType === 'multipart/form-data') {
		var _formParams = this.normalizeParams(formParams);
		for (var key in _formParams) {
			if (_formParams.hasOwnProperty(key)) {
				if (this.isFileParam(_formParams[key])) {
					// file field
					request.attach(key, _formParams[key]);
				} else {
					request.field(key, _formParams[key]);
				}
			}
		}
	} else if (bodyParam) {
		request.send(bodyParam);
	}

	var accept = this.jsonPreferredMime(accepts);
	if (accept) {
		request.accept(accept);
	}

	// TODO If response is 401 or OAuth token has expired, refresh the token if we can?

	request.end(function(error, response) {
		if (callback) {
			var data = null;
			if (!error) {
				data = _this.deserialize(response, returnType);
			}
			callback(error, data, response);
		}
	});

	return request;
};

Client.parseDate = function parseDate(str) {
	str = str.replace(/T/i, ' ');
	return new Date(str);
};

Client.convertToType = function convertToType(data, type) {
	switch (type) {
		case 'Boolean':
			return Boolean(data);
		case 'Integer':
			return parseInt(data, 10);
		case 'Number':
			return parseFloat(data);
		case 'String':
			return String(data);
		case 'Date':
			return this.parseDate(String(data));
		default:
			if (type === Object) {
				// generic object, return directly
				return data;
			} else if (typeof type === 'function') {
				// for model type like: User
				var model = type.constructFromObject(data);
				return model;
			} else if (Array.isArray(type)) {
				// for array type like: ['String']
				var itemType = type[0];
				return data.map(function(item) {
					return Client.convertToType(item, itemType);
				});
			} else if (typeof type === 'object') {
				// for plain object type like: {'String': 'Integer'}
				var keyType, valueType;
				for (var k in type) {
					if (type.hasOwnProperty(k)) {
						keyType = k;
						valueType = type[k];
						break;
					}
				}
				var result = {};
				for (var d in data) {
					if (data.hasOwnProperty(d)) {
						var key = Client.convertToType(d, keyType);
						var value = Client.convertToType(data[d], valueType);
						result[key] = value;
					}
				}
				return result;
			} else {
				// for unknown type, return the data directly
				return data;
			}
	}
};

Client.default = new Client();

module.exports = Client;
